package hr.fer.zemris.ais;

import hr.fer.zemris.graphics.tsp.PrepareTSP;
import hr.fer.zemris.tsp.City;
import hr.fer.zemris.tsp.TSPSolution;
import hr.fer.zemris.tsp.TSPSolutionPool;
import hr.fer.zemris.tsp.TSPUtil;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Random;

/**
 * Razred prikazuje implementaciju algoritma ClonAlg
 * na problemu trgovačkog putnika.
 * 
 * @author marcupic
 */
public class ClonAlg {

	// Polje gradova
	private City[] cities;

	// Parametar koji određuje veličinu populacije klonova
	private int paramBeta;

	// broj novih rješenja (antitijela) koje ćemo dodavati
	private int paramD;
	
	// Broj rješenja u populaciji (broj antitijela)
	private int paramN;
	
	// priručna memorija s rješenjima
	private TSPSolutionPool pool;
	
	// Populacija rješenja
	private TSPSolution[] population;
	
	// Populacija klonova
	private TSPSolution[] clonedPopulation;
	
	// Generator slučajnih brojeva
	private Random rand;
	
	// Veličina populacije klonova
	private int clonedPopulationSize;
	
	// Rangovi u populaciji klonova 
	private int[] clonedPopulationRanks;
	
	/**
	 * Konstruktor.
	 * 
	 * @param cities lista gradova
	 */
	public ClonAlg(List<City> cities) {
		this.cities = new City[cities.size()];
		cities.toArray(this.cities);
		paramN = 100;
		paramD = 10;
		paramBeta = 10;
		this.pool = new TSPSolutionPool(this.cities.length);
		rand = new Random();
		population = new TSPSolution[paramN];
		clonedPopulationSize = 0;
		for(int i = 1; i <= paramN; i++) {
			clonedPopulationSize += (int)((paramBeta*paramN)/((double)i)+0.5);
		}
		clonedPopulation = new TSPSolution[clonedPopulationSize];
		clonedPopulationRanks = new int[clonedPopulationSize];
		initialize();
	}

	/**
	 * Inicijalizacija rješenja (antitijela).
	 */
	private void initialize() {
		for(int i = 0; i < paramN; i++) {
			TSPSolution s = pool.get();
			population[i] = s;
			TSPUtil.randomInitializeSolution(s, rand);
		}
	}

	/**
	 * Glavna petlja optimizacijskog algoritma.
	 */
	public void go() {

		// Postavi parametre
		int iter = 0;
		int iterLimit = 2000;
		
		// Ponavljaj zadani broj puta
		while(iter < iterLimit) {
			iter++;
			TSPUtil.evaluate(population, cities);
			cloning();
			hyperMutation();
			TSPUtil.evaluate(clonedPopulation, cities);
			select();
			birthAndReplace();
		}
		// Najbolje rješenje je prvo zbog sortiranja!
		System.out.println("Best length: "+population[0].tourLength);
		System.out.println(population[0]);
		PrepareTSP.visualize(TSPUtil.reorderCities(cities, population[0].cityIndexes));
	}
	
	/**
	 * Operator kloniranja koji broj klonova koje će stvoriti utvrđuje proporcionalno
	 * dobroti samog rješenja i time generira populaciju klonova.
	 */
	private void cloning() {
		Arrays.sort(population, TSPUtil.solComparator);
		int index = 0;
		for(int i = 1; i <= population.length; i++) {
			TSPSolution s = population[i-1];
			int copies = (int)((paramBeta*paramN)/((double)i)+0.5);
			for(int j = 0; j < copies; j++) {
				TSPSolution c = pool.get();
				System.arraycopy(s.cityIndexes, 0, c.cityIndexes, 0, s.cityIndexes.length);
				clonedPopulation[index] = c;
				clonedPopulationRanks[index] = i;
				index++;
			}
		}
	}

	/**
	 * Operator hipermutacije koji broj mutacija u jediniki ima obrnuto proporcionalan
	 * dobroti jedinke. Što je jedinka bolja, to će se manje mutirati. Sama mutacija
	 * provodi se tako da se zamijeni redosljed dva slučajno odabrana grada.
	 */
	private void hyperMutation() {
		// Kada dodem na zadnju jedinku, zelim obaviti 1+c.cityIndexes.length/4 mutacija
		double tau = 3.476 * (population.length-1);
		// Idem od 1 kako bih najbolje rjesenje, i to njegovu prvu kopiju, ostavio netaknuto
		for(int index = 1; index < clonedPopulation.length; index++) {
			TSPSolution c = clonedPopulation[index];
			int rank = clonedPopulationRanks[index]-1;
			int mutations = (int)(1+c.cityIndexes.length*0.25*(1-Math.exp(-rank/tau))+0.5);
			for(int attempt = 0; attempt < mutations; attempt++) {
				int a = rand.nextInt(c.cityIndexes.length);
				int b = rand.nextInt(c.cityIndexes.length);
				if(a==b) {
					if(b==c.cityIndexes.length-1) {
						b--;
					} else {
						b++;
					}
				}
				int tmp = c.cityIndexes[a];
				c.cityIndexes[a] = c.cityIndexes[b];
				c.cityIndexes[b] = tmp;
			}
		}
	}

	/**
	 * Operator selekcije. Radi nad populacijom klonova i iz nje odabire
	 * najbolja rješenja za novu populaciju. Preostala neiskorištena rješenja
	 * vraćaju se priručnoj memoriji.
	 */
	private void select() {
		Arrays.sort(clonedPopulation, TSPUtil.solComparator);
		for(int i = 0; i < population.length; i++) {
			pool.release(population[i]);
			population[i] = clonedPopulation[i];
		}
		for(int i = population.length; i < clonedPopulation.length; i++) {
			pool.release(clonedPopulation[i]);
		}
	}

	/**
	 * Operator rađanja - D najgorih rješenja nanovo slučajno generira.
	 */
	private void birthAndReplace() {
		int offset = population.length-paramD;
		for(int i=0; i < paramD; i++) {
			TSPUtil.randomInitializeSolution(population[offset+i], rand);
		}
	}

	
	/**
	 * Ulazna točka u program.
	 * 
	 * @param args argumenti komandne linije
	 */
	public static void main(String[] args) throws IOException {
		String fileName = args.length<1 ? 
				"data/gradovi01.txt"
				: args[0];
		List<City> cities = TSPUtil.loadCities(fileName);
		if(cities==null) return;
		new ClonAlg(cities).go();
	}
}
