package hr.fer.zemris.ais;

import hr.fer.zemris.graphics.tsp.PrepareTSP;
import hr.fer.zemris.tsp.City;
import hr.fer.zemris.tsp.TSPSolution;
import hr.fer.zemris.tsp.TSPSolutionPool;
import hr.fer.zemris.tsp.TSPUtil;

import java.io.IOException;
import java.util.List;
import java.util.Random;

/**
 * Razred prikazuje implementaciju algoritma SimpleIA (jednostavan
 * imunološki algoritam) na problemu trgovačkog putnika.
 * 
 * @author marcupic
 */
public class SimpleIA {

	// Polje gradova
	private City[] cities;

	// veličina populacije (broj antitijela)
	private int paramD;
	
	// broj klonova svakog rješenja (antitijela)
	private int paramDup;
	
	// priručna memorija s rješenjima
	private TSPSolutionPool pool;
	
	// Populacija rješenja
	private TSPSolution[] population;
	
	// Populacija klonova
	private TSPSolution[] clonedPopulation;

	// Unija obiju populacija
	private TSPSolution[] unionPopulation;
	
	// Generator slučajnih brojeva
	private Random rand;

	/**
	 * Konstruktor.
	 * 
	 * @param cities lista gradova
	 */
	public SimpleIA(List<City> cities) {
		this.cities = new City[cities.size()];
		cities.toArray(this.cities);
		paramD = 50;
		paramDup = 30;
		this.pool = new TSPSolutionPool(this.cities.length);
		rand = new Random();
		population = new TSPSolution[paramD];
		clonedPopulation = new TSPSolution[paramD*paramDup];
		unionPopulation = new TSPSolution[paramD*paramDup + paramD];
		initialize();
		TSPUtil.evaluate(population, this.cities);
	}

	/**
	 * Inicijalizacija rješenja (antitijela).
	 */
	private void initialize() {
		for(int i = 0; i < paramD; i++) {
			TSPSolution s = pool.get();
			population[i] = s;
			TSPUtil.randomInitializeSolution(s, rand);
		}
	}

	/**
	 * Glavna petlja optimizacijskog algoritma.
	 */
	public void go() {

		// Postavi parametre
		int iter = 0;
		int iterLimit = 2000;
		
		// Ponavljaj zadani broj puta
		while(iter < iterLimit) {
			iter++;
			cloning();
			hyperMutation();
			TSPUtil.evaluate(clonedPopulation, cities);
			select();
		}
		// Najbolje rješenje je prvo zbog sortiranja!
		System.out.println("Best length: "+population[0].tourLength);
		System.out.println(population[0]);
		PrepareTSP.visualize(TSPUtil.reorderCities(cities, population[0].cityIndexes));
	}
	
	/**
	 * Operator kloniranja. Za svaku jedinku iz glavne populacije stvara
	 * zadani broj klonova i time generira populaciju klonova.
	 */
	private void cloning() {
		int index = 0;
		for(int i = 0; i < population.length; i++) {
			TSPSolution s = population[i];
			for(int j = 0; j < paramDup; j++) {
				TSPSolution c = pool.get();
				System.arraycopy(
						s.cityIndexes, 0, c.cityIndexes, 0, s.cityIndexes.length);
				clonedPopulation[index] = c;
				index++;
			}
		}
	}

	/**
	 * Operator hipermutacije. Svaku jedinku iz populacije klonova mutira
	 * tako da zamijeni redosljed dva slučajno odabrana grada.
	 */
	private void hyperMutation() {
		for(int index = 0; index < clonedPopulation.length; index++) {
			TSPSolution c = clonedPopulation[index];
			int a = rand.nextInt(c.cityIndexes.length);
			int b = rand.nextInt(c.cityIndexes.length);
			if(a==b) {
				if(b==c.cityIndexes.length-1) {
					b--;
				} else {
					b++;
				}
			}
			int tmp = c.cityIndexes[a];
			c.cityIndexes[a] = c.cityIndexes[b];
			c.cityIndexes[b] = tmp;
		}
	}

	/**
	 * Operator selekcije. U uniju dodaje izvornu populaciju i klonove,
	 * traži najbolja rješenja za novu populaciju a ostalo otpušta u
	 * priručnu memoriju.
	 */
	private void select() {
		int index = 0;
		for(int i = 0; i < population.length; i++) {
			unionPopulation[index++] = population[i];
		}
		for(int i = 0; i < clonedPopulation.length; i++) {
			unionPopulation[index++] = clonedPopulation[i];
		}
		TSPUtil.partialSort(unionPopulation, population.length);
		for(int i = 0; i < population.length; i++) {
			population[i] = unionPopulation[i];
		}
		for(int i = population.length; i < unionPopulation.length; i++) {
			pool.release(unionPopulation[i]);
		}
	}

	/**
	 * Ulazna točka u program.
	 * 
	 * @param args argumenti komandne linije
	 */
	public static void main(String[] args) throws IOException {
		String fileName = args.length<1 ? 
				"data/gradovi03.txt" 
				: args[0];
		List<City> cities = TSPUtil.loadCities(fileName);
		if(cities==null) return;
		new SimpleIA(cities).go();
	}
}
