package hr.fer.zemris.ga;

import hr.fer.zemris.numeric.IFunkcija;

import java.util.Arrays;
import java.util.Random;

/**
 * Primjer implementacije genetskog algoritma za optimizaciju funkcije jedne
 * varijable.
 *
 * Program koristi binarni prikaz kromosoma.
 * 
 * @author marcupic
 */
public class GeneticAlgorithm {

	/**
	 * Glavni program.
	 * 
	 * @param args argumenti komandne linije - ne koriste se.
	 */
	public static void main(String[] args) {
		
		// Definiranje osnovnih parametara algoritma
		int VEL_POP = 50;
		double VJER_KRIZ = 0.7;
		double VJER_MUT = 0.005;

		// Generator slučajnih brojeva
		Random rand = new Random();

		// Stvaranje dekodera binarnog kromosoma:
		// jedna varijabla, 10 bitova, raspon pretraživanja [-5, 5]
		KromosomDekoder dekoder = new KromosomDekoder(1, 10, -5, 5);

		// Stvaranje dviju populacija n-ta, i (n+1)-va
		// Zbog optimizacije, kasnije ćemo samo mijenjati ta polja
		Kromosom[] populacija = stvoriPopulaciju(VEL_POP, dekoder, rand);
		Kromosom[] novaGeneracija = stvoriPopulaciju(VEL_POP, dekoder, null);

		// Definiranje funkcije koju optimiramo
		IFunkcija funkcija = new IFunkcija() {
			@Override
			public double izracunaj(double[] varijable) {
				int n = varijable.length;
				double vrijednost = 10*n;
				for(int i = 0; i < n; i++) {
					vrijednost += varijable[i]*varijable[i] - 10*Math.cos(2*Math.PI*varijable[i]);
				}
				return vrijednost;
			}
		};
		
		// Početna evaluacija populacije
		evaluirajPopulaciju(populacija, funkcija);

		// Ponovi kroz 1000 generacija
		for(int generacija = 0; generacija < 1000; generacija++) {

			// Sortiraj populaciju po dobroti; najbolja jedinka bit će prva
			Arrays.sort(populacija);
			
			// U novu populaciju prekopiraj dvije najbolje (elitizam!)
			kopiraj(populacija[0], novaGeneracija[0]);
			kopiraj(populacija[1], novaGeneracija[1]);

			// Stvori preostale jedinke nove generacije
			for(int i=1; i < VEL_POP/2; i++) {
				Kromosom roditelj1 = odaberiRoditelja(populacija, rand);
				Kromosom roditelj2 = odaberiRoditelja(populacija, rand);
				Kromosom dijete1 = novaGeneracija[2*i];
				Kromosom dijete2 = novaGeneracija[2*i+1];
				krizaj1TockaPrijeloma(VJER_KRIZ, roditelj1, roditelj2, dijete1, dijete2, rand);
				mutiraj(VJER_MUT, dijete1, rand);
				mutiraj(VJER_MUT, dijete2, rand);
			}
			
			// Zamijeni staru i novu populaciju
			Kromosom[] pomocni = populacija;
			populacija = novaGeneracija;
			novaGeneracija = pomocni;
			
			// Vrednuj populaciju
			evaluirajPopulaciju(populacija, funkcija);
			
			// pronađi najbolje rješenje
			Kromosom najbolji = null;
			for(int i = 0; i < populacija.length; i++) {
				if(i==0 || najbolji.fitnes>populacija[i].fitnes) {
					najbolji = populacija[i];
				}
			}
			// I ispiši ga...
			System.out.println("Trenutno rjesenje: f("+Arrays.toString(najbolji.varijable)
					+") = "+funkcija.izracunaj(najbolji.varijable));
		}
	}

	/**
	 * Pomoćna funkcija koja obavlja kopiranje jednog kromosoma u drugi.
	 * @param original
	 * @param kopija
	 */
	private static void kopiraj(Kromosom original, Kromosom kopija) {
		for(int i = 0; i < original.bitovi.length; i++) {
			kopija.bitovi[i] = original.bitovi[i];
		}
	}

	/**
	 * Stvaranje nove populacije.
	 * 
	 * @param brojJedinki broj jedinki koji treba stvoriti
	 * @param dekoder koji se dekoder koristi
	 * @param rand generator slučajnih brojeva
	 * @return novu populaciju
	 */
	public static Kromosom[] stvoriPopulaciju(int brojJedinki, KromosomDekoder dekoder, Random rand) {
		Kromosom[] populacija = new Kromosom[brojJedinki];
		for(int i = 0; i < populacija.length; i++) {
			if(rand==null) {
				populacija[i] = new Kromosom(dekoder);
			} else {
				populacija[i] = new Kromosom(dekoder, rand);
			}
		}
		return populacija;
	}

	
	/**
	 * Metoda vrednuje predanu populaciju.
	 * 
	 * @param populacija populacija
	 * @param funkcija funkcija koja se optimira
	 */
	private static void evaluirajPopulaciju(Kromosom[] populacija, IFunkcija funkcija) {
		for(int i = 0; i < populacija.length; i++) {
			evaluirajJedinku(populacija[i], funkcija);
		}
	}

	/**
	 * Metoda vrednuje predani kromosom.
	 * 
	 * @param kromosom kromosom
	 * @param funkcija funkcija koja se optimira
	 */
	private static void evaluirajJedinku(Kromosom kromosom, IFunkcija funkcija) {
		kromosom.dekoder.dekodirajKromosom(kromosom);
		kromosom.fitnes = funkcija.izracunaj(kromosom.varijable);
	}

	/**
	 * Metoda za odabir jednog roditelja, gdje je vjerojatnost odabira 
	 * proporcionalna dobroti.
	 * 
	 * @param populacija populacija iz koje se bira
	 * @param rand generator slučajnih brojeva
	 * @return odabranog roditelja
	 */
	private static Kromosom odaberiRoditelja(Kromosom[] populacija, Random rand) {
		double sumaDobrota = 0;
		double najvecaVrijednost = 0;
		for(int i = 0; i < populacija.length; i++) {
			sumaDobrota += populacija[i].fitnes;
			if(i==0 || najvecaVrijednost<populacija[i].fitnes) {
				najvecaVrijednost = populacija[i].fitnes;
			}
		}
		sumaDobrota = populacija.length * najvecaVrijednost - sumaDobrota; 
		double slucajniBroj = rand.nextDouble() * sumaDobrota;
		double akumuliranaSuma = 0;
		for(int i = 0; i < populacija.length; i++) {
			akumuliranaSuma += najvecaVrijednost - populacija[i].fitnes;
			if(slucajniBroj<akumuliranaSuma) return populacija[i];
		}
		return populacija[populacija.length-1];
	}

	/**
	 * Metoda obavlja križanje s jednom točkom prijeloma. Križanje se obavlja s
	 * zadanom vjerojatnošću. Ako se ne dogodi križanje, kao djeca se vraćaju
	 * roditelji.
	 * 
	 * @param vjerKriz vjerojatnost križanja (decimalni broj između 0 i 1)
	 * @param roditelj1 prvi roditelj
	 * @param roditelj2 drugi roditelj
	 * @param dijete1 prvo dijete
	 * @param dijete2 drugo dijete
	 * @param rand generator slučajnih brojeva
	 */
	private static void krizaj1TockaPrijeloma(double vjerKriz, Kromosom roditelj1, 
			Kromosom roditelj2, Kromosom dijete1, Kromosom dijete2, Random rand) {
		if(rand.nextFloat() <= vjerKriz) {
			int tockaPrijeloma = rand.nextInt(roditelj1.dekoder.ukupnoBitova-1)+1;
			for(int i = 0; i < tockaPrijeloma; i++) {
				dijete1.bitovi[i] = roditelj1.bitovi[i];
				dijete2.bitovi[i] = roditelj2.bitovi[i];
			}
			for(int i = tockaPrijeloma; i < roditelj1.dekoder.ukupnoBitova; i++) {
				dijete1.bitovi[i] = roditelj2.bitovi[i];
				dijete2.bitovi[i] = roditelj1.bitovi[i];
			}
		} else {
			for(int i = 0; i < roditelj1.dekoder.ukupnoBitova; i++) {
				dijete1.bitovi[i] = roditelj1.bitovi[i];
				dijete2.bitovi[i] = roditelj2.bitovi[i];
			}
		}
	}

	/**
	 * Operator mutacije. Bitovi se okreću zadanom vjerojatnošću.
	 * 
	 * @param vjerMut vjerojatnost mutacije bita (broj od 0 do 1)
	 * @param dijete dijete koje se mutira
	 * @param rand generator slučajnih brojeva
	 */
	private static void mutiraj(double vjerMut, Kromosom dijete, Random rand) {
		for(int i = 0; i < dijete.dekoder.ukupnoBitova; i++) {
			if(rand.nextFloat() <= vjerMut) {
				dijete.bitovi[i] = (byte)(1-dijete.bitovi[i]);
			}
		}
	}

}
