package hr.fer.zemris.ga;

import java.util.Random;

/**
 * Binarni kromosom - jedno rješenje genetskog algoritma. Pretpostavlja
 * se da je posrijedi riješavanje problema koji se sastoji od više realnih
 * varijabli što se reflektira u građi samog kromosoma.
 * 
 * Važno: ugrađena funkcija za usporedbu kromosoma pretpostavlja
 * da se radi o minimizacijskom problemu.
 * 
 * @author marcupic
 */
public class Kromosom implements Comparable<Kromosom> {
	// Bitovi kromosoma
	byte[] bitovi;
	// Vrijednost funkcije dobrote kromosoma (zapravo, to je vrijednost
	// funkcije u promatranoj točki)
	double fitnes;
	// Vrijednosti realnih varijabli koje kromosom predstavlja 
	double[] varijable;
	// Dekoder koji zna konvertirati binarni prikaz u realne varijable
	KromosomDekoder dekoder;
	
	/**
	 * Konstruktor koji stvara novi kromosom ali ga ne inicijalizira.
	 * 
	 * @param dekoder dekoder kromosoma
	 */
	public Kromosom(KromosomDekoder dekoder) {
		this.dekoder = dekoder;
		this.bitovi = new byte[dekoder.ukupnoBitova];
		this.fitnes = 0;
		this.varijable = new double[dekoder.brojVarijabli];
	}

	/**
	 * Konstruktor koji stvara novi kromosom i inicijalizira
	 * ga na slučajni uzorak bitova.
	 * 
	 * @param dekoder dekoder kromosoma
	 * @param rand generator slučajnih brojeva
	 */
	public Kromosom(KromosomDekoder dekoder, Random rand) {
		this.dekoder = dekoder;
		this.bitovi = new byte[dekoder.ukupnoBitova];
		this.fitnes = 0;
		this.varijable = new double[dekoder.brojVarijabli];
		for(int i = 0; i < dekoder.ukupnoBitova; i++) {
			this.bitovi[i] = rand.nextBoolean() ? (byte)1 : (byte)0;
		}
	}
	
	/**
	 * Funkcija za definiranje prirodnog poretka kromosoma. Pretpostavka
	 * je da se radi minimizacijski problem pa je manji (bolji) onaj kromosom
	 * koji ima manju vrijednost {@linkplain #fitnes} koja zapravo čuva vrijednost
	 * funkcije u promatranoj točki.
	 *  
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	@Override
	public int compareTo(Kromosom o) {
		if(this.fitnes < o.fitnes) {
			return -1;
		}
		if(this.fitnes > o.fitnes) {
			return 1;
		}
		return 0;
	}
}
