package hr.fer.zemris.graphics.tsp;

import hr.fer.zemris.tsp.City;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.WindowConstants;

/**
 * Razred koji obavlja vizualizaciju pronađene rute kod problema TSP.
 * 
 * @author marcupic
 */
public class PrepareTSP extends JFrame {

	private static final long serialVersionUID = 1L;

	// Lista gradova; poredak već predstavlja rutu
	private List<City> cities = new ArrayList<City>();
	// Pomoćna labela za prikaz duljine rute
	private JLabel duljinaLabel;
	// Pomoćna labela za prikaz brja gradova
	private JLabel brojLabel;

	// Pomoćna komponenta koja obavlja iscrtavanje slike gradova
	private VisualizeComponent komponenta;

	/**
	 * Konstruktor koji prima polje gradova. Redosljed elemenata automatski određuje
	 * i samu turu.
	 * 
	 * @param presetCities polje gradova
	 */
	public PrepareTSP(City[] presetCities) {
		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		setSize(500, 500);
		setLocation(20, 20);
	
		if(presetCities!=null) {
			for(City c : presetCities) {
				cities.add(c);
			}
		}
		komponenta = new VisualizeComponent();
		this.getContentPane().setLayout(new BorderLayout());
		this.getContentPane().add(komponenta, BorderLayout.CENTER);

		JPanel p = new JPanel(new GridLayout(1,2));
		duljinaLabel = new JLabel();
		brojLabel = new JLabel();
		p.add(brojLabel);
		p.add(duljinaLabel);
		this.getContentPane().add(p, BorderLayout.PAGE_END);
		
		azurirajDuljinu();
		setVisible(true);
	}

	/**
	 * Pomoćna funkcija koja računa duljinu ture i ažurira podatke u labelama
	 * {@linkplain #duljinaLabel} i {@linkplain #brojLabel}.
	 */
	private void azurirajDuljinu() {
		double d = 0;
		if(!cities.isEmpty()) {
			City cc = null;
			City oldCc = null;
			for(int i = 0; i < cities.size(); i++) {
				cc = cities.get(i);
				if(oldCc!=null) {
					d += Math.sqrt((oldCc.x-cc.x)*(oldCc.x-cc.x) + (oldCc.y-cc.y)*(oldCc.y-cc.y));
				}
				oldCc = cc;
			}
			cc = cities.get(0);
			if(cc!=oldCc) {
				d += Math.sqrt((oldCc.x-cc.x)*(oldCc.x-cc.x) + (oldCc.y-cc.y)*(oldCc.y-cc.y));
			}
		}
		duljinaLabel.setText("Duljina: "+((double)((int)(d*1000+0.5))/1000.0));
		brojLabel.setText("Broj gradova: "+cities.size());
		
	}

	/**
	 * Pomoćna metoda koju smije pozvati proizvoljna dretva a služi za inicijalizaciju
	 * prikaza ture. Metoda prima polje gradova i generira prikaz. Polje se pri tome 
	 * kopira pa je dretva-pozivatelj slobodna kasnije obavljati modifikacije nad poljem;
	 * to više neće imati nikakvog utjecaja na prikaz.
	 *  
	 * @param cities polje gradova čijim je poretkom u polju ujedno određena i tura
	 */
	public static void visualize(final City[] cities) {
		try {
			SwingUtilities.invokeAndWait(new Runnable() {
				@Override
				public void run() {
					new PrepareTSP(cities);
				}
			});
		} catch (Exception ignorable) {
			System.out.println("Prikaz nije moguć.");
		}
	}
	
	/**
	 * Pomoćna komponenta koja crta turu.
	 * 
	 * @author marcupic
	 */
	private class VisualizeComponent extends JComponent {

		private static final long serialVersionUID = 1L;

		@Override
		protected void paintComponent(Graphics g) {
			super.paintComponent(g);
			if(cities.isEmpty()) return;
			City cc = null;
			City oldCc = null;
			g.setColor(Color.BLACK);
			for(int i = 0; i < cities.size(); i++) {
				cc = cities.get(i);
				if(oldCc!=null) {
					g.drawLine(oldCc.x, oldCc.y, cc.x, cc.y);
				}
				oldCc = cc;
			}
			cc = cities.get(0);
			if(cc!=oldCc) {
				g.drawLine(oldCc.x, oldCc.y, cc.x, cc.y);
			}
			g.setColor(Color.BLACK);
			for(int i = 0; i < cities.size(); i++) {
				cc = cities.get(i);
				g.fillRect(cc.x-2, cc.y-2, 6, 6);
			}
		}
	}
}
