package hr.fer.zemris.pso;

import hr.fer.zemris.numeric.IFunkcija;

import java.text.DecimalFormat;
import java.util.Random;

/**
 * Primjer uporabe algoritma roja čestica za optimizaciju.
 * 
 * @author marcupic
 */
public class ParticleSwarmOptimization {

	// Funkcija koju optimiramo
	private IFunkcija funkcija;
	
	// Mimimumi varijabli u prostoru pretraživanja
	private double[] varMin;
	
	// Maksimumi varijabli u prostoru pretraživanja
	private double[] varMax;

	// Maksimalna promjene varijable odjednom, za svaku dimenziju
	private double[] velBounds;

	// Postotak raspona prostora pretraživanja koji se koristi za 
	// izračun ograničenja pomaka u jednom koraku 
	double velBoundsPercentage;

	// Globalni brojač iteracija
	private int iteracija;
	
	// Težina koju korisimo na početku
	private double linWeightStart;

	// Težina koju korisimo na kraju smanjivanja
	private double linWeightEnd;

	// Iteracija u kojoj težina pada na krajnju i dalje se ne mijenja
	private int linWeightTreshold;

	// Broj čestica s kojima radimo
	int VEL_POP;
	
	// Generator slučajnih brojeva
	Random rand;
	
	// Broj dimenzija funkcije 
	int dims;

	// Konstanta c1
	double c1;

	// Konstanta c2
	double c2;
	
	// Čestice
	Particle[] particles;
	
	// Susjedstvo
	Neighborhood neighborhood;

	// Pomoć u formatiranu brojeva
	DecimalFormat df = new DecimalFormat("0.00000");

	/**
	 * Glavni program.
	 * 
	 * @param args argumenti komandne linije - ne koriste se.
	 */
	public static void main(String[] args) throws Exception {
		final ParticleSwarmOptimization t = new ParticleSwarmOptimization();
		t.go();
	}

	/**
	 * Konstruktor.
	 */
	public ParticleSwarmOptimization() {
		
		// Definiranje parametara algoritma
		VEL_POP = 20;
		dims = 2;
		c1 = 2;
		c2 = 2.5;

		iteracija = 0;
		linWeightStart = 0.9;
		linWeightEnd = 0.4;
		linWeightTreshold = 50;

		// Definiranje minimalnih i maksimalnih vrijednosti po dimenzijama,
		// te definiranje maksimalne promjene varijable u jednom koraku.
		varMin = new double[dims];
		varMax = new double[dims];
		velBounds = new double[dims];
		velBoundsPercentage = 0.05;  // dozvoli pomak od 5% raspona
		for(int d = 0; d < dims; d++) {
			varMin[d] = -5;
			varMax[d] =  5;
			velBounds[d] = (varMax[d]-varMin[d])*velBoundsPercentage;
		}

		// Definiranje funkcije koju optimiramo.
		funkcija = new IFunkcija() {
			@Override
			public double izracunaj(double[] varijable) {
				int n = varijable.length;
				double vrijednost = 10*n;
				for(int i = 0; i < n; i++) {
					vrijednost += varijable[i]*varijable[i] - 10*Math.cos(2*Math.PI*varijable[i]);
				}
				return vrijednost;
			}
		};

		// Generator slučajnih brojeva.
		rand = new Random();
		
		// Inicijalizacija
		particles = new Particle[VEL_POP];
		for(int i = 0; i < VEL_POP; i++) {
			particles[i] = new Particle(dims);
			for(int d = 0; d < dims; d++) {
				particles[i].vars[d] = rand.nextDouble()*(varMax[d]-varMin[d])+varMin[d];
				particles[i].oldVars[d] = particles[i].vars[d];
				particles[i].bestVars[d] = particles[i].vars[d];
				particles[i].velocity[d] = rand.nextDouble()*(2*velBounds[d])-velBounds[d];
			}
			particles[i].bestValue = funkcija.izracunaj(particles[i].vars);
		}

		// Definiranje susjedstva koje koristimo - lokalno veličine 5
		neighborhood = new LocalNeighborhood(VEL_POP, dims, 5, true);
		
		// Za odabir globalnog susjedstva može se iskoristiti sljedeće: 
		//neighborhood = new GlobalNeighborhood(VEL_POP, dims, true);

		// Ispiši inicijalnu statistiku.
		updateStatistics();
	}

	/**
	 * Glavna metoda optimizacijskog algoritma.
	 */
	public void go() {
		// Ponavljaj zadani broj puta
		for(int iter = 0; iter < 100; iter++) {
			nextIteration();
		}
	}

	/**
	 * Jedna iteracija algoritma PSO.
	 */
	protected void nextIteration() {
		iteracija++;
		
		// Koju težinu koristimo? Težina linearno pada s iteracijama do
		// neke male zadane, i dalje ostaje konstantna.
		double w;
		if(iteracija>linWeightTreshold) {
			w = linWeightEnd;
		} else {
			w = linWeightStart + (linWeightEnd-linWeightStart)*(iteracija-1.0)/linWeightTreshold;
		}
		
		// Ažurirajmo "znanje" susjedstva
		neighborhood.scan(particles);
		
		// Ažuriraj pozicije i brzine svake čestice
		for(int i = 0; i < particles.length; i++) {
			double[] socialBest = neighborhood.findBest(i);
			for(int d = 0; d < dims; d++) {
				particles[i].oldVars[d] = particles[i].vars[d];
				particles[i].velocity[d] = 
					w * particles[i].velocity[d] 
					+ c1*rand.nextDouble()*(particles[i].bestVars[d]-particles[i].vars[d])
					+ c2*rand.nextDouble()*(socialBest[d]-particles[i].vars[d])
					;
				if(particles[i].velocity[d] < -velBounds[d]) {
					particles[i].velocity[d] = -velBounds[d];
				} else if(particles[i].velocity[d] > velBounds[d]) {
					particles[i].velocity[d] = velBounds[d];
				}
				particles[i].vars[d] = particles[i].vars[d] + particles[i].velocity[d];
			}
		}
		
		// Izračunaj vrijednost funkcije u novim pozicijama svih čestica,
		// i po potrebi ažuriraj najbolje rješenje čestice
		for(int i = 0; i < particles.length; i++) {
			particles[i].value = funkcija.izracunaj(particles[i].vars);
			if(particles[i].value < particles[i].bestValue) {
				particles[i].bestValue = particles[i].value;
				for(int d = 0; d < dims; d++) {
					particles[i].bestVars[d] = particles[i].vars[d];
				}
			}
		}
	
		// Ispiši statistiku na ekran
		updateStatistics();
	}

	/**
	 * Pomoćna metode koja na ekran ispisuje statističke podatke o populaciji,
	 * te najbolje pronađeno rješenje.
	 */
	private void updateStatistics() {
		int bestindex = 0;
		double bestValue = particles[bestindex].bestValue;
		double sum = bestValue;
		for(int i = 1; i < particles.length; i++) {
			if(particles[i].bestValue<bestValue) {
				bestValue = particles[i].bestValue;
				bestindex = i;
			}
			sum += particles[i].bestValue;
		}
		StringBuilder sb = new StringBuilder();
		sb.append("Iter: ");
		sb.append(iteracija);
		sb.append(", Average: ");
		sb.append(df.format(sum/particles.length));
		sb.append(", (");
		for(int d = 0; d < dims; d++) {
			if(d>0) {
				sb.append(", ");
			}
			sb.append(df.format(particles[bestindex].bestVars[d]));
		}
		sb.append(")=");
		sb.append(df.format(bestValue));
		System.out.println(sb.toString());
	}

}
