package hr.fer.zemris.tsp;

import hr.fer.zemris.util.ArraysUtil;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Random;

/**
 * Pomoćni razred koji sadrži metode vezane uz TSP.
 * 
 * @author marcupic
 */
public class TSPUtil {

	/**
	 * Metoda učitava listu gradova iz datoteke. Datoteka je tekstualna.
	 * U svakom retku nalazi se x i y koordinata grada razdvojene znakom
	 * tab.
	 * 
	 * @param fileName naziv datoteke
	 * @return listu gradova ili null ako dođe do pogreške
	 * @throws IOException ako se dogodi pogreška u radu s datotekom
	 */
	public static List<City> loadCities(String fileName) throws IOException {
		BufferedReader br = null;
		try {
			br = new BufferedReader(new FileReader(fileName));
			List<City> cities = new ArrayList<City>();
			while(true) {
				String line = br.readLine();
				if(line==null) break;
				line = line.trim();
				if(line.isEmpty()) continue;
				String[] elems = line.split("\\t");
				cities.add(new City(Integer.parseInt(elems[0]),Integer.parseInt(elems[1])));
			}
			br.close();
			return cities;
		} catch(IOException ex) {
			System.out.println("Pogreška prilikom rada s datotekom "+fileName);
			if(br!=null) try { br.close(); } catch(Exception ignorable) {}
			return null;
		}
	}
	
	/**
	 * Metoda koja služi parcijalnom sortiranju predanog polja rješenja TSP-a.
	 * Zadatak metode je na početak polja staviti <code>number</code> najboljih
	 * rješenja (to su ona s najmanjom duljinom ture); poredak preostalog dijela
	 * polja nije bitan.
	 * 
	 * @param population rješenja koja treba parcijalno sortirati
	 * @param number broj najboljih rješenja koja treba staviti na početak polja
	 */
	public static void partialSort(TSPSolution[] population, int number) {
		for(int i = 0; i < number; i++) {
			int best = i;
			for(int j = i+1; j < population.length; j++) {
				if(population[best].tourLength > population[j].tourLength) {
					best = j;
				}
			}
			if(best != i) {
				TSPSolution tmp = population[i];
				population[i] = population[best];
				population[best] = tmp;
			}
		}
	}

	/**
	 * Metoda za predano rješenje računa njegovu duljinu temeljem predane
	 * matrice udaljenosti. U toj matrici, na mjestu [i,j] nalazi se udaljenost
	 * od grada <code>i</code> do grada <code>j</code>.
	 * 
	 * @param sol rješenje za koje treba izračunati duljinu ture
	 * @param distanceMatrix matrica udaljenosti gradova
	 */
	public static void evaluate(TSPSolution sol, double[][] distanceMatrix) {
		int pocetni = sol.cityIndexes[0];
		double distance = 0;
		for(int i = 1; i < sol.cityIndexes.length; i++) {
			distance += distanceMatrix[pocetni][sol.cityIndexes[i]];
			pocetni = sol.cityIndexes[i];
		}
		distance += distanceMatrix[pocetni][sol.cityIndexes[0]];
		sol.tourLength = distance;
	}

	/**
	 * Metoda za predano rješenje računa njegovu duljinu temeljem predanog
	 * polja gradova i indeksa koji se nalaze u samom rješenju.
	 * 
	 * @param sol rješenje
	 * @param cities polje gradova
	 */
	public static void evaluate(TSPSolution sol, City[] cities) {
		int pocetni = sol.cityIndexes[0];
		double distance = 0;
		for(int i = 1; i < sol.cityIndexes.length; i++) {
			City a = cities[pocetni]; 
			City b = cities[sol.cityIndexes[i]];
			distance += Math.sqrt((a.x-b.x)*(a.x-b.x)+(a.y-b.y)*(a.y-b.y));
			pocetni = sol.cityIndexes[i];
		}
		City a = cities[pocetni]; 
		City b = cities[sol.cityIndexes[0]];
		distance += Math.sqrt((a.x-b.x)*(a.x-b.x)+(a.y-b.y)*(a.y-b.y));
		sol.tourLength = distance;
	}

	/**
	 * Pomoćna metoda koja računa dobrotu svih tura u predanom polju.
	 * 
	 * @param population polje rješenja
	 * @param cities polje gradova
	 */
	public static void evaluate(TSPSolution[] population, City[] cities) {
		for(int i = 0; i < population.length; i++) {
			TSPSolution s = population[i];
			evaluate(s, cities);
		}
	}

	/**
	 * Metoda vraća novo polje gradova ne temelju izvornog polja gradova
	 * i predanog redosljeda određenog indeksima.
	 * 
	 * @param cities originalno polje gradova
	 * @param indexes željeni poredak gradova
	 * @return novo polje složeno prema indeksima
	 */
	public static City[] reorderCities(City[] cities, int[] indexes) {
		City[] array = new City[indexes.length];
		for(int i = 0; i < indexes.length; i++) {
			City c = cities[indexes[i]];
			array[i] = new City(c.x, c.y);
		}
		return array;
	}
	
	/**
	 * Pomoćna metoda koja stvara nasumični poredak gradova.
	 * 
	 * @param sol rješenje
	 * @param rand generator slučajnih brojeva
	 */
	public static void randomInitializeSolution(TSPSolution sol, Random rand) {
		ArraysUtil.linearFillArray(sol.cityIndexes);
		ArraysUtil.shuffleArray(sol.cityIndexes, rand);
	}

	/**
	 * Komparator dvaju rješenja. Rješenje je manje ako je duljina ture manja.
	 */
	public static Comparator<TSPSolution> solComparator = new Comparator<TSPSolution>() {
		@Override
		public int compare(TSPSolution o1, TSPSolution o2) {
			double razlika = o1.tourLength - o2.tourLength;
			return razlika < 0 ? -1 : (razlika > 0 ? 1 : 0);
		}
	};

}
